@@ -43,7 +43,7 @@ def get_available_applications(as_subprocess=False):
             import subprocess
             params = {"env": env, "stdout": subprocess.PIPE, "stderr": subprocess.PIPE}
             with subprocess.Popen(cmd_args, **params) as p:
-                logger.debug("{} '{}'".format(' '.join(cmd_args[:-1]), pycmd))
+                logger.debug(f"{' '.join(cmd_args[:-1])} '{pycmd}'")
                 stdout, stderr = p.communicate()
                 stdout, stderr = stdout.decode(), stderr.decode()
                 # ast.literal_eval is secure and will raise more handy Exceptions than eval
@@ -54,7 +54,7 @@ def get_available_applications(as_subprocess=False):
         except subprocess.SubprocessError:
             logger.debug("Failed to call subprocess")
         except (ValueError, SyntaxError, AssertionError):
-            logger.debug("Failed to decode output or convert to tuple :" + "\nstdout={}\nstderr={}".format(stdout, stderr))
+            logger.debug("Failed to decode output or convert to tuple :" + f"\nstdout={stdout}\nstderr={stderr}")
         if not app_list:
             logger.info("Failed to list applications in an independent process. Falling back to local otb import")
 
@@ -64,7 +64,7 @@ def get_available_applications(as_subprocess=False):
         logger.warning("Unable to load applications. Set env variable OTB_APPLICATION_PATH then try again")
         return ()
 
-    logger.info("Successfully loaded {} OTB applications".format(len(app_list)))
+    logger.info(f"Successfully loaded {len(app_list)} OTB applications")
     return app_list
 
 
@@ -83,6 +83,26 @@ class {name}(App):
     def __init__(self, *args, **kwargs):
         super().__init__('{name}', *args, **kwargs)
 """
-# Here we could customize the template and overwrite special methods depending on application
 for _app in AVAILABLE_APPLICATIONS:
+    # Default behavior for any OTB application
     exec(_code_template.format(name=_app))
+
+    # Customize the behavior for TensorflowModelServe application. The user doesn't need to set the env variable
+    # `OTB_TF_NSOURCES`, it is handled in pyotb
+    if _app == 'TensorflowModelServe':
+        class TensorflowModelServe(App):
+            def set_nb_sources(self, *args, n_sources=None):
+                """
+                Set the number of sources of TensorflowModelServe. Can be either user-defined or deduced from the args
+                """
+                if n_sources:
+                    os.environ['OTB_TF_NSOURCES'] = str(int(n_sources))
+                else:
+                    # Retrieving the number of `source#.il` parameters
+                    params_dic = {k: v for arg in args if isinstance(arg, dict) for k, v in arg.items()}
+                    n_sources = len([k for k in params_dic if k.startswith('source') and k.endswith('.il')])
+                    os.environ['OTB_TF_NSOURCES'] = str(n_sources)
+
+            def __init__(self, *args, n_sources=None, **kwargs):
+                self.set_nb_sources(*args, n_sources=n_sources)
+                super().__init__('TensorflowModelServe', *args, **kwargs)
