@@ -52,7 +52,7 @@ class otbObject(ABC):
             if isinstance(arg, dict):
                 kwargs.update(arg)
             elif isinstance(arg, str) and kwargs:
-                logger.warning('{}: Keyword arguments specified, ignoring argument "{}"'.format(self.name, arg))
+                logger.warning(f'{self.name}: Keyword arguments specified, ignoring argument "{arg}"')
             elif isinstance(arg, str):
                 if hasattr(self, 'output_parameter_key'):  # this is for Output, Operation...
                     output_parameter_key = self.output_parameter_key
@@ -63,14 +63,14 @@ class otbObject(ABC):
         # Handling pixel types
         pixel_types = {}
         if isinstance(pixel_type, str):  # this correspond to 'uint8' etc...
-            pixel_type = getattr(otb, 'ImagePixelType_{}'.format(pixel_type))
+            pixel_type = getattr(otb, f'ImagePixelType_{pixel_type}')
             pixel_types = {param_key: pixel_type for param_key in kwargs}
         elif isinstance(pixel_type, int):  # this corresponds to ImagePixelType_...
             pixel_types = {param_key: pixel_type for param_key in kwargs}
         elif isinstance(pixel_type, dict):  # this is to specify a different pixel type for each output
             for key, this_pixel_type in pixel_type.items():
                 if isinstance(this_pixel_type, str):
-                    this_pixel_type = getattr(otb, 'ImagePixelType_{}'.format(this_pixel_type))
+                    this_pixel_type = getattr(otb, f'ImagePixelType_{this_pixel_type}')
                 if isinstance(this_pixel_type, int):
                     pixel_types[key] = this_pixel_type
 
@@ -85,8 +85,8 @@ class otbObject(ABC):
                     if not out_fn.endswith('?'):
                         out_fn += "?"
                     out_fn += filename_extension
-                    logger.info('{}: Using extended filename for output.'.format(self.name))
-                logger.info('{}: write output file "{}" to {}'.format(self.name, output_parameter_key, out_fn))
+                    logger.info(f'{self.name}: Using extended filename for output.')
+                logger.info(f'{self.name}: write output file "{output_parameter_key}" to {out_fn}')
                 self.app.SetParameterString(output_parameter_key, out_fn)
 
                 if output_parameter_key in pixel_types:
@@ -114,7 +114,7 @@ class otbObject(ABC):
         elif isinstance(key, tuple) and len(key) == 2:
             # adding a 3rd dimension
             key = key + (slice(None, None, None),)
-        (rows, cols, channels) = key
+        (cols, rows, channels) = key
         return Slicer(self, rows, cols, channels)
 
     def __getattr__(self, name):
@@ -231,16 +231,32 @@ class otbObject(ABC):
     def __hash__(self):
         return id(self)
 
-    def __array__(self):
+    def to_numpy(self, propagate_pixel_type=False):
         """
-        This is called when running np.asarray(pyotb_object)
+        Export a pyotb object to numpy array
         :return: a numpy array
         """
         if hasattr(self, 'output_parameter_key'):  # this is for Input, Output, Operation, Slicer
             output_parameter_key = self.output_parameter_key
         else:  # this is for App
             output_parameter_key = self.output_parameters_keys[0]
-        return self.app.ExportImage(output_parameter_key)['array']
+        # we make a copy to avoid some segfault if the reference to app is lost
+        array = self.app.ExportImage(output_parameter_key)['array'].copy()
+        if propagate_pixel_type:
+            otb_pixeltype = get_pixel_type(self)
+            otb_pixeltype_to_np_pixeltype = {0: np.uint8, 1: np.int16, 2: np.uint16, 3: np.int32, 4: np.uint32,
+                                             5: np.float32, 6: np.float64}
+            np_pixeltype = otb_pixeltype_to_np_pixeltype[otb_pixeltype]
+            array = array.astype(np_pixeltype)
+        return array
+
+    def __array__(self):
+        """
+        This is called when running np.asarray(pyotb_object)
+        :return: a numpy array
+        """
+
+        return self.to_numpy()
 
     def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
         """
@@ -303,7 +319,7 @@ class Slicer(otbObject):
         :param channels:
         """
         # Initialize the app that will be used for writing the slicer
-        app = App('ExtractROI', {"in": input, 'mode': 'extent'})
+        app = App('ExtractROI', {"in": input, 'mode': 'extent'}, propagate_pixel_type=True)
         self.output_parameter_key = 'out'
         self.name = 'Slicer'
 
@@ -322,12 +338,12 @@ class Slicer(otbObject):
             elif isinstance(channels, tuple):
                 channels = list(channels)
             elif not isinstance(channels, list):
-                raise ValueError('Invalid type for channels, should be int, slice or list of bands. : {}'.format(channels))
+                raise ValueError(f'Invalid type for channels, should be int, slice or list of bands. : {channels}')
 
             # Change the potential negative index values to reverse index
             channels = [c if c >= 0 else nb_channels + c for c in channels]
 
-            app.set_parameters(cl=['Channel{}'.format(i + 1) for i in channels])
+            app.set_parameters(cl=[f'Channel{i + 1}' for i in channels])
 
         # Spatial slicing
         spatial_slicing = False
@@ -363,10 +379,10 @@ class Input(otbObject):
     Class for transforming a filepath to pyOTB object
     """
     def __init__(self, filepath):
-        self.app = App('ExtractROI', filepath).app
+        self.app = App('ExtractROI', filepath, propagate_pixel_type=True).app
         self.output_parameter_key = 'out'
         self.filepath = filepath
-        self.name = 'Input from {}'.format(filepath)
+        self.name = f'Input from {filepath}'
 
     def __str__(self):
         return '<pyotb.Input object from {}>'.format(self.filepath)
@@ -379,7 +395,7 @@ class Output(otbObject):
     def __init__(self, app, output_parameter_key):
         self.app = app  # keeping a reference of the OTB app
         self.output_parameter_key = output_parameter_key
-        self.name = 'Output {} from {}'.format(output_parameter_key, self.app.GetName())
+        self.name = f'Output {output_parameter_key} from {self.app.GetName()}'
 
     def __str__(self):
         return '<pyotb.Output {} object, id {}>'.format(self.app.GetName(), id(self))
@@ -412,7 +428,7 @@ class App(otbObject):
         # This will only store if app has been excuted, then find_output() is called when accessing the property
         self._finished = val
 
-    def __init__(self, appname, *args, execute=True, image_dic=None, otb_stdout=True, **kwargs):
+    def __init__(self, appname, *args, execute=True, image_dic=None, otb_stdout=True, propagate_pixel_type=False, **kwargs):
         """
         Enables to run an otb app as a oneliner. Handles in-memory connection between apps
         :param appname: name of the app, e.g. 'Smoothing'
@@ -426,6 +442,9 @@ class App(otbObject):
                           the result of app.ExportImage(). Use it when the app takes a numpy array as input.
                           See this related issue for why it is necessary to keep reference of object:
                           https://gitlab.orfeo-toolbox.org/orfeotoolbox/otb/-/issues/1824
+        :param otb_stdout: whether to print logs of the app
+        :param propagate_pixel_type: Propagate the pixel type from inputs to output. If several inputs, the type of an
+                                     arbitrary input is considered. If several outputs, all will have the same type.
         :param kwargs: keyword arguments e.g. il=['input1.tif', App_object2, App_object3.out], out='output.tif'
         """
         self.appname = appname
@@ -441,11 +460,13 @@ class App(otbObject):
         if args or kwargs:
             self.set_parameters(*args, **kwargs)
         else:
-            logger.warning("{}: No parameters where provided. Use App.set_parameters() then App.execute()".format(self.name))
+            logger.warning(f"{self.name}: No parameters where provided. Use App.set_parameters() then App.execute()")
             execute = False
         # Run app, write output if needed, update `finished` property
         if execute:
             self.execute()
+        if propagate_pixel_type:
+            self.__propagate_pixel_type()
         # 'Saving' outputs as attributes, i.e. so that they can be accessed like that: App.out
         # Also, thanks to __getitem__ method, the outputs can be accessed as App["out"]. This is useful when the key
         # contains reserved characters such as a point eg "io.out"
@@ -458,15 +479,15 @@ class App(otbObject):
         Execute with appropriate App method and outputs to disk if needed"
         :return: boolean flag that indicate if command executed with success
         """
-        logger.debug("{}: run execute() with parameters={}".format(self.name, self.parameters))
+        logger.debug(f"{self.name}: run execute() with parameters={self.parameters}")
         try:
             if self.__with_output():
                 self.app.ExecuteAndWriteOutput()
             self.app.Execute()
             self.finished = True
         except (RuntimeError, FileNotFoundError) as e:
-            raise Exception('{}: error during during app execution'.format(self.name)) from e
-        logger.debug("{}: execution succeeded".format(self.name))
+            raise Exception(f'{self.name}: error during during app execution') from e
+        logger.debug(f"{self.name}: execution succeeded")
         return self.finished
 
     def find_output(self):
@@ -486,7 +507,7 @@ class App(otbObject):
                 missing.append(filename)
         if missing:
             for filename in missing:
-                logger.error("{}: execution seems to have failed, {} does not exist".format(self.name, filename))
+                logger.error(f"{self.name}: execution seems to have failed, {filename} does not exist")
                 #raise FileNotFoundError(filename)
         return files
 
@@ -517,23 +538,26 @@ class App(otbObject):
         :param kwargs: keyword arguments e.g. il=['input1.tif', oApp_object2, App_object3.out], out='output.tif'
         :return: boolean flag that indicate if app was correctly set using given parameters
         """
-        self.parameters.update(self.__parse_args(args))
-        self.parameters.update(kwargs)
+        parameters = kwargs
+        parameters.update(self.__parse_args(args))
         # Going through all arguments
-        for param, obj in self.parameters.items():
+        for param, obj in parameters.items():
             if param not in self.app.GetParametersKeys():
-                raise Exception("{}: parameter '{}' was not recognized. ".format(self.name, param) +
-                                "Available keys are {}".format(self.app.GetParametersKeys()))
+                raise Exception(f"{self.name}: parameter '{param}' was not recognized. "
+                                f"Available keys are {self.app.GetParametersKeys()}")
             # When the parameter expects a list, if needed, change the value to list
             if self.__is_key_list(param) and not isinstance(obj, (list, tuple)):
-                self.parameters[param] = [obj]
+                parameters[param] = [obj]
                 obj = [obj]
             try:
                 # This is when we actually call self.app.SetParameter*
                 self.__set_param(param, obj)
             except (RuntimeError, TypeError, ValueError) as e:
-                raise Exception("{}: something went wrong before execution ".format(self.name) +
-                                "(while setting parameter {} to '{}')".format(param, obj)) from e
+                raise Exception(f"{self.name}: something went wrong before execution "
+                                f"(while setting parameter {param} to '{obj}')") from e
+
+        # update the attribute
+        self.parameters.update(parameters)
 
     # Private functions
     @staticmethod
@@ -584,6 +608,21 @@ class App(otbObject):
         else:
             self.app.SetParameterValue(param, obj)
 
+    def __propagate_pixel_type(self):
+        """Propagate the pixel type from inputs to output. If several inputs, the type of an arbitrary input
+           is considered. If several outputs, all outputs will have the same type."""
+        pixel_type = None
+        for param in self.parameters.values():
+            try:
+                pixel_type = get_pixel_type(param)
+            except TypeError:
+                pass
+        if pixel_type is None:  # we use this syntax because pixel_type can be equal to 0
+            logger.warning(f"{self.name}: Could not propagate pixel type from inputs to output, " +
+                           f"no valid input found")
+        else:
+            for out_key in self.output_parameters_keys:
+                self.app.SetParameterOutputImagePixelType(out_key, pixel_type)
 
     def __with_output(self):
         """Check if App has any output parameter key"""
@@ -608,7 +647,7 @@ class App(otbObject):
 
     # Special methods
     def __str__(self):
-        return '<pyotb.App {} object id {}>'.format(self.appname, id(self))
+        return f'<pyotb.App {self.appname} object id {id(self)}>'
 
 
 class Operation(otbObject):
@@ -673,7 +712,7 @@ class Operation(otbObject):
         else:
             self.app = App('BandMathX', il=self.unique_inputs, exp=self.exp).app
 
-        self.name = 'Operation exp="{}"'.format(self.exp)
+        self.name = f'Operation exp="{self.exp}"'
 
     def create_fake_exp(self, operator, inputs, nb_bands=None):
         """
@@ -687,7 +726,7 @@ class Operation(otbObject):
         self.inputs = []
         self.nb_channels = {}
 
-        logger.debug("{}, {}".format(operator, inputs))
+        logger.debug(f"{operator}, {inputs}")
         # this is when we use the ternary operator with `pyotb.where` function. The output nb of bands is already known
         if operator == '?' and nb_bands:
             nb_bands = nb_bands
@@ -731,13 +770,13 @@ class Operation(otbObject):
 
             # Generating the fake expression of the whole operation
             if len(inputs) == 1:  # this is only for 'abs'
-                fake_exp = '({}({}))'.format(operator, fake_exps[0])
+                fake_exp = f'({operator}({fake_exps[0]}))'
             elif len(inputs) == 2:
                 # We create here the "fake" expression. For example, for a BandMathX expression such as '2 * im1 + im2',
                 # the false expression stores the expression 2 * str(input1) + str(input2)
-                fake_exp = '({} {} {})'.format(fake_exps[0], operator, fake_exps[1])
+                fake_exp = f'({fake_exps[0]} {operator} {fake_exps[1]})'
             elif len(inputs) == 3 and operator == '?':  # this is only for ternary expression
-                fake_exp = '({} ? {} : {})'.format(fake_exps[0], fake_exps[1], fake_exps[2])
+                fake_exp = f'({fake_exps[0]} ? {fake_exps[1]} : {fake_exps[2]})'
 
             self.fake_exp_bands.append(fake_exp)
 
@@ -764,7 +803,7 @@ class Operation(otbObject):
                 nb_channels = input.input.nb_channels
             else:
                 # Add the band number (e.g. replace '<pyotb.App object>' by '<pyotb.App object>b1')
-                fake_exp = str(input.input) + 'b{}'.format(input.one_band_sliced)
+                fake_exp = str(input.input) + f'b{input.one_band_sliced}'
                 inputs = [input.input]
                 nb_channels = {input.input: 1}
         # For logicalOperation, we save almost the same attributes as an Operation
@@ -786,7 +825,7 @@ class Operation(otbObject):
             nb_channels = {input: get_nbchannels(input)}
             inputs = [input]
             # Add the band number (e.g. replace '<pyotb.App object>' by '<pyotb.App object>b1')
-            fake_exp = str(input) + 'b{}'.format(band)
+            fake_exp = str(input) + f'b{band}'
 
         return fake_exp, inputs, nb_channels
 
@@ -807,7 +846,7 @@ class Operation(otbObject):
         return exp_bands, exp
 
     def __str__(self):
-        return '<pyotb.Operation `{}` object, id {}>'.format(self.operator, id(self))
+        return f'<pyotb.Operation `{self.operator}` object, id {id(self)}>'
 
 
 class logicalOperation(Operation):
@@ -854,12 +893,12 @@ class logicalOperation(Operation):
 
             # We create here the "fake" expression. For example, for a BandMathX expression such as 'im1 > im2',
             # the logical fake expression stores the expression "str(input1) > str(input2)"
-            logical_fake_exp = '({} {} {})'.format(fake_exps[0], operator, fake_exps[1])
+            logical_fake_exp = f'({fake_exps[0]} {operator} {fake_exps[1]})'
 
             # We keep the logical expression, useful if later combined with other logical operations
             self.logical_fake_exp_bands.append(logical_fake_exp)
             # We create a valid BandMath expression, e.g. "str(input1) > str(input2) ? 1 : 0"
-            fake_exp = '({} ? 1 : 0)'.format(logical_fake_exp)
+            fake_exp = f'({logical_fake_exp} ? 1 : 0)'
             self.fake_exp_bands.append(fake_exp)
 
 
@@ -876,10 +915,8 @@ def get_nbchannels(inp):
         try:
             info = App("ReadImageInfo", inp, otb_stdout=False)
             nb_channels = info.GetParameterInt("numberbands")
-        except (RuntimeError, TypeError) as e:
-            logger.error('Not a valid image : {}'.format(inp))
-            logger.error('{}'.format(e))
-            nb_channels = None
+        except Exception:  # this happens when we pass a str that is not a filepath
+            raise TypeError(f'Could not get the number of channels of `{inp}`. Not a filepath or wrong filepath')
     return nb_channels
 
 
@@ -887,26 +924,27 @@ def get_pixel_type(inp):
     """
     Get the encoding of input image pixels
     :param inp: a filepath, or any pyotb object
-    :return pixel_type: either a dict of pixel types (in case of an App where there are several outputs)
-                        format is like `otbApplication.ImagePixelType_uint8'
+    :return pixel_type: format is like `otbApplication.ImagePixelType_uint8', which actually is an int. For an App
+                        with several outputs, only the pixel type of the first output is returned
+
     """
     if isinstance(inp, str):
         # Executing the app, without printing its log
-        info = App("ReadImageInfo", inp, otb_stdout=False)
-        datatype = info.GetParameterInt("datatype")  # which is such as short, float...
-        dataype_to_pixeltype = {
-            'unsigned_char': 'uint8', 'short': 'int16', 'unsigned_short': 'uint16', 
-            'int': 'int32', 'unsigned_int': 'uint32', 'long': 'int32', 'ulong': 'uint32', 
-            'float': 'float','double': 'double'
-        }
+        try:
+            info = App("ReadImageInfo", inp, otb_stdout=False)
+        except Exception:  # this happens when we pass a str that is not a filepath
+            raise TypeError(f'Could not get the pixel type of `{inp}`. Not a filepath or wrong filepath')
+        datatype = info.GetParameterString("datatype")  # which is such as short, float...
+        dataype_to_pixeltype = {'unsigned_char': 'uint8', 'short': 'int16', 'unsigned_short': 'uint16',
+                                'int': 'int32', 'unsigned_int': 'uint32', 'long': 'int32', 'ulong': 'uint32',
+                                'float': 'float', 'double': 'double'}
         pixel_type = dataype_to_pixeltype[datatype]
-        pixel_type = getattr(otb, 'ImagePixelType_{}'.format(pixel_type))
-    elif isinstance(inp, (Input, Output, Operation)):
-        pixel_type = inp.GetImageBasePixelType(inp.output_parameter_key)
+        pixel_type = getattr(otb, f'ImagePixelType_{pixel_type}')
+    elif isinstance(inp, (Input, Output, Operation, Slicer)):
+        pixel_type = inp.GetParameterOutputImagePixelType(inp.output_parameter_key)
     elif isinstance(inp, App):
-        if len(inp.output_parameters_keys) > 1:
-            pixel_type = inp.GetImageBasePixelType(inp.output_parameters_keys[0])
-        else:
-            pixel_type = {key: inp.GetImageBasePixelType(key) for key in inp.output_parameters_keys}
+        pixel_type = inp.GetParameterOutputImagePixelType(inp.output_parameters_keys[0])
+    else:
+        raise TypeError(f'Could not get the pixel type. Not supported type: {inp}')
 
     return pixel_type
