@@ -1,8 +1,10 @@
 # -*- coding: utf-8 -*-
+import inspect
 import os
+import sys
+import textwrap
 import uuid
 import logging
-import multiprocessing
 from collections import Counter
 
 from .core import (App, Input, Operation, logicalOperation, get_nbchannels)
@@ -34,7 +36,7 @@ def where(cond, x, y):
     if x_nb_channels and y_nb_channels:
         if x_nb_channels != y_nb_channels:
             raise Exception('X and Y images do not have the same number of bands. ' +
-                            'X has {} bands whereas Y has {} bands'.format(x_nb_channels, y_nb_channels))
+                            f'X has {x_nb_channels} bands whereas Y has {y_nb_channels} bands')
 
     x_or_y_nb_channels = x_nb_channels if x_nb_channels else y_nb_channels
     cond_nb_channels = get_nbchannels(cond)
@@ -46,12 +48,12 @@ def where(cond, x, y):
         out_nb_channels = cond_nb_channels
 
     if cond_nb_channels != 1 and x_or_y_nb_channels and cond_nb_channels != x_or_y_nb_channels:
-        raise Exception('Condition and X&Y do not have the same number of bands. Condition has ' +
+        raise Exception(f'Condition and X&Y do not have the same number of bands. Condition has '
                         '{cond_nb_channels} bands whereas X&Y have {x_or_y_nb_channels} bands')
 
     # If needed, duplicate the single band binary mask to multiband to match the dimensions of x & y
     if cond_nb_channels == 1 and x_or_y_nb_channels and x_or_y_nb_channels != 1:
-        logger.info('The condition has one channel whereas X/Y has/have {} channels. Expanding number of channels '.format(x_or_y_nb_channels) +
+        logger.info(f'The condition has one channel whereas X/Y has/have {x_or_y_nb_channels} channels. Expanding number of channels '
                     'of condition to match the number of channels or X/Y')
 
     operation = Operation('?', cond, x, y, nb_bands=out_nb_channels)
@@ -235,7 +237,7 @@ def define_processing_area(*args, window_rule='intersection', pixel_size_rule='m
             # TODO : it is when the user wants the final bounding box to be the union of all bounding box
             #  It should replace any 'outside' pixel by some NoData -> add `fillvalue` argument in the function
 
-        logger.info('Cropping all images to extent Upper Left ({}, {}), Lower Right ({}, {})'.format(ULX, ULY, LRX, LRY))
+        logger.info(f'Cropping all images to extent Upper Left ({ULX}, {ULY}), Lower Right ({LRX}, {LRY})')
 
         # Applying this bounding box to all inputs
         new_inputs = []
@@ -252,7 +254,7 @@ def define_processing_area(*args, window_rule='intersection', pixel_size_rule='m
                     reference_pixel_size_input = new_input
             except Exception as e:
                 logger.error(e)
-                logger.error('Images may not intersect : {}'.format(input))
+                logger.error(f'Images may not intersect : {input}')
                 # TODO: what should we do then? return an empty raster ? fail ? return None ?
         inputs = new_inputs
 
@@ -279,7 +281,7 @@ def define_processing_area(*args, window_rule='intersection', pixel_size_rule='m
             pass
             # TODO : when the user explicitely specify the pixel size -> add argument inside the function
         pixel_size = metadatas[reference_input]['GeoTransform'][1]
-        logger.info('Resampling all inputs to resolution : {}'.format(pixel_size))
+        logger.info(f'Resampling all inputs to resolution : {pixel_size}')
 
         # Perform resampling on inputs that do not comply with the target pixel size
         new_inputs = []
@@ -321,7 +323,7 @@ def define_processing_area(*args, window_rule='intersection', pixel_size_rule='m
 
 def run_tf_function(func):
     """
-    This decorator enables using a function that calls some TF operations, with pyotb object as inputs.
+    This function enables using a function that calls some TF operations, with pyotb object as inputs.
 
     For example, you can write a function that uses TF operations like this :
         @run_tf_function
@@ -335,68 +337,98 @@ def run_tf_function(func):
     :param func: function taking one or several inputs and returning *one* output
     :return wrapper: a function that returns a pyotb object
     """
+    try:
+        from .apps import TensorflowModelServe
+    except ImportError:
+        raise Exception('Could not run Tensorflow function: failed to import TensorflowModelServe. Check that you '
+                        'have OTBTF configured (https://github.com/remicres/otbtf#how-to-install)')
 
-    def create_and_save_tf_model(output_dir, *inputs):
+    def get_tf_pycmd(output_dir, channels, scalar_inputs):
         """
-        Simply creates the TF model and save it to temporary location.
-        //!\\ Currently incompatible with OTBTF, to be run inside a multiprocessing.Process
-        //!\\ Does not work if OTBTF has been used previously in the script
+        Create a string containing all python instructions necessary to create and save the Keras model
 
         :param output_dir: directory under which to save the model
-        :param inputs: a list of pyotb objects or int/float
+        :param channels: list of raster channels (int). Contain `None` entries for non-raster inputs
+        :param scalar_inputs: list of scalars (int/float). Contain `None` entries for non-scalar inputs
         """
-        import tensorflow as tf
-        # Change the raster inputs to TF inputs
-        model_inputs = []  # model inputs corresponding to rasters
-        tf_inputs = []  # inputs for the TF function corresponding to all inputs (rasters, int...)
-        for input in inputs:
-            if not isinstance(input, (int, float)):
-                nb_bands = input.shape[-1]
-                input = tf.keras.Input((None, None, nb_bands))
-                model_inputs.append(input)
-            tf_inputs.append(input)
 
-        # call the TF operations on theses inputs
-        output = func(*tf_inputs)
+        # Getting the string definition of the tf function (e.g. "def multiply(x1, x2):...")
+        # TODO: maybe not entirely foolproof, maybe we should use dill instead? but it would add a dependency
+        func_def_str = inspect.getsource(func)
+        func_name = func.__name__
 
-        # Create and save the .pb model
-        model = tf.keras.Model(inputs=model_inputs, outputs=output)
-        model.save(output_dir)
+        create_and_save_model_str = func_def_str
+
+        # Adding the instructions to create the model and save it to output dir
+        create_and_save_model_str += textwrap.dedent(f"""
+            import tensorflow as tf
+
+            model_inputs = []
+            tf_inputs = []
+            for channel, scalar_input in zip({channels}, {scalar_inputs}):
+                if channel:
+                    input = tf.keras.Input((None, None, channel))
+                    tf_inputs.append(input)
+                    model_inputs.append(input)
+                else:
+                    if isinstance(scalar_input, int):  # TF doesn't like mixing float and int
+                        scalar_input = float(scalar_input)
+                    tf_inputs.append(scalar_input)
+
+            output = {func_name}(*tf_inputs)
+            
+            # Create and save the .pb model
+            model = tf.keras.Model(inputs=model_inputs, outputs=output)
+            model.save("{output_dir}")
+            """)
+
+        return create_and_save_model_str
 
     def wrapper(*inputs, tmp_dir='/tmp'):
         """
-        For the user point of view, this function simply applies some TF operations to some rasters.
-        Underlyingly, it saveq a .pb model that describe the TF operations, then creates an OTB ModelServe application
-        that applies this .pb model to the actual inputs.
+        For the user point of view, this function simply applies some TensorFlow operations to some rasters.
+        Underlyingly, it saves a .pb model that describe the TF operations, then creates an OTB ModelServe application
+        that applies this .pb model to the inputs.
 
         :param inputs: a list of pyotb objects, filepaths or int/float numbers
         :param tmp_dir: directory where temporary models can be written
         :return: a pyotb object, output of TensorFlowModelServe
         """
-        # Change potential string filepaths to pyotb objects
-        inputs = [Input(input) if isinstance(input, str) and not isinstance(input, (int, float)) else input for input in
-                  inputs]
+        # Get infos about the inputs
+        channels = []
+        scalar_inputs = []
+        raster_inputs = []
+        for input in inputs:
+            try:
+                # this is for raster input
+                channel = get_nbchannels(input)
+                channels.append(channel)
+                scalar_inputs.append(None)
+                raster_inputs.append(input)
+            except Exception:
+                # this is for other inputs (float, int)
+                channels.append(None)
+                scalar_inputs.append(input)
 
-        # Create and save the model. This is executed **inside an independent process** because (as of 2021-11),
+        # Create and save the model. This is executed **inside an independent process** because (as of 2022-03),
         # tensorflow python library and OTBTF are incompatible
-        out_savedmodel = os.path.join(tmp_dir, 'tmp_otbtf_model_{}'.format(uuid.uuid4()))
-        p = multiprocessing.Process(target=create_and_save_tf_model, args=(out_savedmodel, *inputs,))
-        p.start()
-        p.join()
-
-        # Getting the nb of inputs and setting it for OTBTF
-        raster_inputs = [input for input in inputs if not isinstance(input, (int, float))]
-        nb_model_inputs = len(raster_inputs)
-        os.environ['OTB_TF_NSOURCES'] = str(nb_model_inputs)
-
-        # Run the OTBTF model serving application
-        model_serve = App('TensorflowModelServe',
-                          {'model.dir': out_savedmodel,
-                           'optim.disabletiling': 'on', 'model.fullyconv': 'on'}, execute=False)
+        out_savedmodel = os.path.join(tmp_dir, f'tmp_otbtf_model_{uuid.uuid4()}')
+        pycmd = get_tf_pycmd(out_savedmodel, channels, scalar_inputs)
+        cmd_args = [sys.executable, "-c", pycmd]
+        try:
+            import subprocess
+            subprocess.run(cmd_args, env=os.environ, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        except subprocess.SubprocessError:
+            logger.debug("Failed to call subprocess")
+        if not os.path.isdir(out_savedmodel):
+            logger.info("Failed to save the model")
 
+        # Initialize the OTBTF model serving application
+        model_serve = TensorflowModelServe({'model.dir': out_savedmodel, 'optim.disabletiling': 'on',
+                                            'model.fullyconv': 'on'}, n_sources=len(raster_inputs), execute=False)
+        # Set parameters and execute
         for i, input in enumerate(raster_inputs):
-            model_serve.set_parameters({'source{}.il'.format(i + 1): [input]})
-
+            model_serve.set_parameters({f'source{i + 1}.il': [input]})
         model_serve.Execute()
         # TODO: handle the deletion of the temporary model ?
 
